# -*- coding: utf-8 -*-

# Copyright 2008-2010 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import re
import os.path
from os.path import isfile, isdir
import time
import cgi
USE_PIL = False
try:
    import gd
except ImportError:
    USE_PIL = True
    from PIL import Image
from skoollib import Udg
from skoolparser import TableParser, RefParser

BLACK = (0, 0, 0)
BLUE = (0, 0, 197)
RED = (197, 0, 0)
MAGENTA = (197, 0, 197)
GREEN = (0, 198, 0)
CYAN = (0, 198, 197)
YELLOW = (197, 198, 0)
WHITE = (205, 198, 205)
BRIGHT_BLACK = (0, 0, 0)
BRIGHT_BLUE = (0, 0, 255)
BRIGHT_RED = (255, 0, 0)
BRIGHT_MAGENTA = (255, 0, 255)
BRIGHT_GREEN = (0, 255, 0)
BRIGHT_CYAN = (0, 255, 255)
BRIGHT_YELLOW = (255, 255, 0)
BRIGHT_WHITE = (255, 255, 255)
TRANSPARENT = (0, 254, 0)

# Ref file section prefixes
GLOSSARY = 'Glossary'
BUG = 'Bug'
FACT = 'Fact'
POKE = 'Poke'

class SkoolWriter:
    def __init__(self, skool=None, file_info=None):
        self.skool = skool
        self.file_info = file_info
        if skool:
            self.table_parser = TableParser()
            self._create_macros()
            if USE_PIL:
                self._pil_create_colours()
        self.time = time.strftime("%a %d %b %Y %H:%M:%S %Z")

    @classmethod
    def write_home_page(cls, topdir):
        games = []
        games.append(('Skool Daze', FileInfo.sd_dir))
        games.append(('Back to Skool', FileInfo.bts_dir))
        ofile = open(join(topdir, FileInfo.home), "w")
        cwd = '..'
        writer = SkoolWriter()
        writer.write_prologue(ofile)
        ofile.write('<head>\n')
        ofile.write('<title>Skool Disassemblies</title>\n')
        ofile.write('<link rel="stylesheet" type="text/css" href="%s" />\n' % FileInfo.css_relpath(cwd))
        ofile.write('</head>\n')
        ofile.write('<body class="main">\n')
        ofile.write('<h1>Skool Disassemblies</h1>\n')
        ofile.write('<ul class="home">\n')
        for game_name, game_dir in games:
            game_index = join(game_dir, FileInfo.index)
            ofile.write('<li><a class="link" href="%s">The complete %s RAM disassembly</a></li>\n' % (game_index, game_name))
        ofile.write('<li><a class="link" href="%s">Changelog</a></li>\n' % FileInfo.changelog)
        ofile.write('<li><a class="link" href="http://pyskool.ca/?page_id=113">Old versions</a></li>\n')
        ofile.write('</ul>\n')
        ofile.write('<div class="copyrights">\n')
        ofile.write('<div class="copyright">&#169; 1984 Microsphere Computer Services Ltd (Skool Daze)</div>\n')
        ofile.write('<div class="copyright">&#169; 1985 Microsphere Computer Services Ltd (Back to Skool)</div>\n')
        ofile.write('<div class="copyright">&#169; 2001, 2008-2010 Richard Dymond (Skool Disassemblies)</div>\n')
        ofile.write('</div>\n')
        writer.write_footer(ofile)

    @classmethod
    def _get_changelog(cls, writer):
        ref_parser = RefParser()
        ref_parser.parse(FileInfo.changelog_ref)
        changelog = []
        for title, paragraphs in ref_parser.get_ref_items('Changelog', True):
            intro = '' if paragraphs[0][0] == '-' else paragraphs[0][0]
            items = []
            for p in paragraphs[1:]:
                items.append([writer.expand_vars(line) for line in p])
            changelog.append((title, intro, items))
        return changelog

    @classmethod
    def write_changelog(cls, topdir):
        ofile = open(join(topdir, FileInfo.changelog), 'w')
        cwd = '..'
        writer = SkoolWriter()
        writer.write_prologue(ofile)
        ofile.write('<head>\n')
        ofile.write('<title>Skool Disassemblies: Changelog</title>\n')
        ofile.write('<link rel="stylesheet" type="text/css" href="%s" />\n' % FileInfo.css_relpath(cwd))
        ofile.write('</head>\n')
        ofile.write('<body class="main">\n')
        ofile.write('<h1>Skool Disassemblies: Changelog</h1>\n')
        changelog = cls._get_changelog(writer)
        ofile.write('<ul class="linkList">\n')
        for title, description, items in changelog:
            ofile.write('<li><a class="link" href="#%s">%s</a></li>\n' % (title, title))
        ofile.write('</ul>\n')
        for title, description, items in changelog:
            ofile.write('<div class="changelog">\n')
            ofile.write('<div class="changelogTitle"><a name="%s">%s</a></div>\n' % (title, title))
            if description:
                ofile.write('<div class="changelogDesc">%s</div>\n' % description)
            if items:
                ofile.write('<ul class="changelog">\n')
                for item in items:
                    indents = [0]
                    for line in item:
                        new_indent = len(line) - len(line.lstrip())
                        if new_indent != indents[-1]:
                            if new_indent > indents[-1]:
                                ofile.write('<ul>\n')
                                indents.append(new_indent)
                            else:
                                while new_indent < indents[-1]:
                                    ofile.write('</ul>\n')
                                    indents.pop()
                        ofile.write('<li>%s</li>\n' % line)
                    while len(indents) > 1:
                        ofile.write('</ul>\n')
                        indents.pop()
                ofile.write('</ul>\n')
            ofile.write('</div>\n')
        writer.write_footer(ofile)

    def has_gbuffer(self):
        return self.skool.has_gbuffer()

    def has_glossary(self):
        return self.skool.has_ref_items(GLOSSARY)

    def has_bugs(self):
        return self.skool.has_ref_items(BUG)

    def has_facts(self):
        return self.skool.has_ref_items(FACT)

    def has_pokes(self):
        return self.skool.has_ref_items(POKE)

    def file_exists(self, fname):
        return self.file_info.file_exists(fname)

    def write_asm_entries(self):
        self.write_entries(FileInfo.asm_path, FileInfo.memory_map)

    def write_start_routine(self):
        start_routine = self.skool.entries.get(str(self.skool.start_address))
        self.write_entry(FileInfo.asm_path, start_routine)

    def write_keytable(self):
        ofile, cwd = self.open_file(FileInfo.keys)
        self.write_header(ofile, "Keys", cwd, "keys")
        keytable_details = self.expand(self.skool.get_section('KeytableDetails'), cwd)
        ofile.write('<div class="details">\n%s\n</div>\n' % keytable_details)
        return ofile, cwd, self.skool.keypress_routines

    def write_bubble(self, ofile, bubble_id, img, title, text):
        ofile.write('<div id="%s" class="B P" onclick="h(this)">' % bubble_id)
        ofile.write('<div class="H">%s</div>' % title)
        ofile.write('<table><tr><td>%s</td><td>%s</td></tr></table>' % (img, text))
        ofile.write('</div>')

    def write_playarea_tile_table(self, ofile, cwd, tiles, id_prefix='s'):
        ofile.write('<table class="tiles">\n')
        for row in tiles:
            ofile.write('<tr>\n')
            for tile in row:
                ofile.write('<td>')
                if tile:
                    bubble_id = '%s%ix%i' % (id_prefix, tile.x, tile.y)
                    title = '(x,y)=(%i,%i)' % (tile.x, tile.y)
                    img = self.expand('#PATILE%i,%i' % (tile.x, tile.y), cwd, tile=tile)
                    ofile.write('<div class="T" onclick="s(\'%s\')">%s</div>' % (bubble_id, img))
                    self.write_bubble(ofile, bubble_id, img, title, '<br/>'.join(self.get_playarea_tile_text(tile)))
                ofile.write('</td>\n')
            ofile.write('</tr>\n')
        ofile.write('</table>\n')

    def write_mutable_tile_tables(self, ofile, cwd):
        pass

    def write_playarea_tiles(self):
        ofile, cwd = self.open_file(FileInfo.playarea_tiles)
        self.write_header(ofile, 'Play area tiles', cwd, 'graphics', js=True)
        info = self.expand_vars(self.skool.get_section('PlayAreaTileInfo'))
        ofile.write(self.expand(info, cwd))
        self.write_playarea_tile_table(ofile, cwd, self.skool.get_skool_udgs(0, 0, self.skool.skool_width, 21))
        self.write_mutable_tile_tables(ofile, cwd)
        self.write_footer(ofile)

    def _write_skool(self, fname, title, sprites=False):
        ofile, cwd = self.open_file(fname)
        self.write_header(ofile, title, cwd, 'graphics')
        img = self.expand('#SKOOL0,0,%i,21,2,%i,8' % (self.skool.skool_width, 1 if sprites else 0), cwd)
        ofile.write('<div>%s</div>\n' % img)
        self.write_footer(ofile)

    def write_playarea(self):
        self._write_skool(FileInfo.playarea, 'Play area')

    def write_asstart(self):
        self._write_skool(FileInfo.asstart, 'Initial animatory states and locations', True)

    def write_animatory_states(self):
        ofile, cwd = self.open_file(FileInfo.animatory_states)
        self.write_header(ofile, 'Animatory states', cwd, 'graphics')
        ofile.write('<table>\n')
        for n in range(128):
            row = []
            ofile.write('<tr>\n')
            for state in (n, n + 128):
                ofile.write('<td>\n')
                ofile.write('<a name="%i" />%s\n' % (state, self.expand('#ASIMG%i' % state, cwd)))
                ofile.write('</td>\n')
            ofile.write('<td>%i, %i: %s</td>\n' % (n, n + 128, self.skool.as_descs[n]))
            ofile.write('</tr>\n')
        ofile.write('</table>\n')
        self.write_footer(ofile)

    def write_animatory_state_tiles(self):
        ofile, cwd = self.open_file(FileInfo.animatory_state_tiles)
        self.write_header(ofile, 'Animatory state tiles', cwd, 'graphics', js=True)

        info = self.expand_vars(self.skool.get_section('AnimatoryStateTileInfo'))
        ofile.write(self.expand(info, cwd))

        attr = 120
        ofile.write('<table>\n')
        for n in range(128):
            row = []
            ofile.write('<tr>\n')
            states = (n, n + 128) if self.skool.game_id == 'sd' else (n,)
            for state in states:
                ofile.write('<td>\n')
                ofile.write('<table>\n')
                sprite = self.skool.build_sprite(state, attr)
                for row_num in range(len(sprite)):
                    ofile.write('<tr>\n')
                    row = sprite[row_num]
                    for col_num in range(len(row)):
                        tile = row[col_num]
                        bubble_id = 'b%i_%ix%i' % (state, row_num, col_num)
                        title = '%i: row %i, column %i' % (state, row_num, col_num)
                        img = self.expand('#ASTILE%i,%i,%i,%i' % (state, row_num, col_num, attr), cwd, tile=tile)
                        text = ['<i>lsb</i>=%i, <i>ref_page</i>=%i' % (tile.ref_addr % 256, tile.ref_addr // 256)]
                        text.append('<i>ref_addr</i>=(<i>lsb</i>,<i>ref_page</i>)=%i' % tile.ref_addr)
                        if tile.ref == 0:
                            text.append('<i>ref</i>=%i' % tile.ref)
                        else:
                            text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
                            text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
                        ofile.write('<td>')
                        ofile.write('<div class="T" onclick="s(\'%s\')">%s</div>' % (bubble_id, img))
                        self.write_bubble(ofile, bubble_id, img, title, '<br/>'.join(text))
                        ofile.write('</td>\n')
                    ofile.write('</tr>\n')
                ofile.write('</table>\n')
                ofile.write('</td>\n')
            ofile.write('<td>%s: %s</td>\n' % (', '.join([str(s) for s in states]), self.skool.as_descs[n]))
            ofile.write('</tr>\n')
        ofile.write('</table>\n')

        self.write_footer(ofile)

    def write_graphics(self):
        ofile, cwd = self.open_file(FileInfo.graphics)
        self.write_header(ofile, 'Graphics', cwd, 'graphics')
        return ofile, cwd

    def write_graphic_glitches(self):
        self.write_box_page(FileInfo.graphic_glitches, 'Graphic glitches', 'graphics', self.skool.get_ref_items('GraphicGlitch'))

    def write_index(self, minimal=False):
        ofile, cwd = self.open_file(FileInfo.index)
        self.write_head(ofile, 'Index', cwd)
        ofile.write('<body class="main">\n')
        table = Table("header")
        row = []
        row.append(Cell("The complete", "headerText"))
        row.append(Cell(self._get_logo(cwd)))
        row.append(Cell('RAM disassembly', "headerText"))
        table.add_row(row)
        ofile.write(table.html())

        sections = []
        sections.append(('<div class="headerText">Memory maps</div>\n', self._get_memory_map_links(cwd, minimal)))
        sections.append(('<div class="headerText">Graphics</div>\n', self._get_graphics_links(cwd, minimal)))
        sections.append(('<div class="headerText">Data tables and buffers</div>\n', self._get_data_links(cwd, minimal)))
        sections.append(('<div class="headerText">Other code</div>\n', self._get_other_code_links(cwd, minimal)))
        sections.append(('<div class="headerText">Reference</div>\n', self._get_reference_links(cwd, minimal)))
        for header, links in sections:
            if links:
                ofile.write(header)
                ofile.write('<ul class="indexList">\n')
                for href, link_text, other_text in links:
                    ofile.write('  <li><a class="link" href="%s">%s</a>%s</li>\n' % (href, link_text, other_text))
                ofile.write('</ul>\n')

        self.write_footer(ofile)

    def add_link(self, links, cwd, fname, link_text, minimal, desc=''):
        if not minimal or self.file_info.file_exists(fname):
            link_file = FileInfo.relpath(cwd, fname)
            links.append((link_file, link_text, desc))

    def _get_memory_map_links(self, cwd, minimal):
        links = []
        self.add_link(links, cwd, FileInfo.memory_map, 'Everything', minimal, ' (routines, data, text and unused addresses)')
        self.add_link(links, cwd, FileInfo.routines, 'Routines', minimal, ' (just the routines)')
        self.add_link(links, cwd, FileInfo.data, 'Data', minimal, ' (just the data)')
        self.add_link(links, cwd, FileInfo.messages, 'Messages', minimal, ' (just the text)')
        return links

    def _get_graphics_links(self, cwd, minimal):
        links = []
        self.add_link(links, cwd, FileInfo.playarea, 'Play area', minimal)
        self.add_link(links, cwd, FileInfo.playarea_tiles, 'Play area tiles', minimal)
        self.add_link(links, cwd, FileInfo.asstart, 'Initial animatory states and locations', minimal)
        self.add_link(links, cwd, FileInfo.animatory_states, 'Animatory states', minimal)
        self.add_link(links, cwd, FileInfo.animatory_state_tiles, 'Animatory state tiles', minimal)
        self.add_link(links, cwd, FileInfo.graphics, 'Other graphics', minimal)
        self.add_link(links, cwd, FileInfo.graphic_glitches, 'Graphic glitches', minimal)
        return links

    def _get_data_links(self, cwd, minimal):
        links = []
        self.add_link(links, cwd, FileInfo.gbuffer, 'Game status buffer', minimal)
        self.add_link(links, cwd, FileInfo.cbuffer, 'Character buffers', minimal)
        ttable_file = FileInfo.asm_relpath('', self.skool.timetable_address)
        self.add_link(links, cwd, ttable_file, 'Main timetable', minimal)
        self.add_link(links, cwd, FileInfo.personal_timetables, 'Personal timetables', minimal)
        self.add_link(links, cwd, FileInfo.lesson_index, 'Lessons', minimal)
        taps_file = FileInfo.asm_relpath('', self.skool.tap_address_table)
        self.add_link(links, cwd, taps_file, 'Command lists', minimal)
        self.add_link(links, cwd, FileInfo.keys, 'Keypress table', minimal)
        return links

    def _get_other_code_links(self, cwd, minimal):
        links = []
        self.add_link(links, cwd, FileInfo.load_index, 'Loading code', minimal)
        self.add_link(links, cwd, FileInfo.save, 'Save code', minimal)
        start_file = FileInfo.asm_relpath('', self.skool.start_address)
        self.add_link(links, cwd, start_file, 'Startup code', minimal)
        return links

    def _get_reference_links(self, cwd, minimal):
        links = []
        self.add_link(links, cwd, FileInfo.glossary, 'Glossary', minimal)
        self.add_link(links, cwd, FileInfo.facts, 'Trivia', minimal)
        self.add_link(links, cwd, FileInfo.bugs, 'Bugs', minimal)
        self.add_link(links, cwd, FileInfo.pokes, 'Pokes', minimal)
        return links

    def write_lessons(self):
        cwd = FileInfo.lessons_path
        index_path = FileInfo.relpath(cwd, FileInfo.lesson_index)
        up = '<a class="link" href="%s">Index</a>' % index_path
        for i in range(self.skool.min_lesson, self.skool.max_lesson + 1):
            ofile = self.file_info.open_file(cwd, FileInfo.lesson_fname(i))
            self.write_header(ofile, "Lesson %i" % i, cwd, "lesson")
            prev = None if i == self.skool.min_lesson else self.expand('#LESSON%i' % (i - 1), cwd)
            next = None if i == self.skool.max_lesson else self.expand('#LESSON%i' % (i + 1), cwd)
            self.write_prev_next(ofile, prev, next, up)
            ofile.write('<div class="description">Lesson %i: %s</div>\n' % (i, self.skool.lesson_descs[i]))
            lesson_table = Table("lesson", 'Character', 'Address', 'Command list')
            for c in range(self.skool.min_character, self.skool.max_character + 1):
                lesson_address = i + 256 * c
                tap = self.skool.snapshot[lesson_address]
                row = []
                pt_link = self.expand('#PT%i' % c, cwd)
                char_desc = self.skool.characters[c]
                row.append(Cell("%s - %s" % (pt_link, char_desc), "lessonCharacter"))
                row.append(Cell(lesson_address, "lessonAddress"))
                tap_link = self.expand('#TAP%i' % tap, cwd)
                tap_desc = self.skool.tap_descs[tap]
                row.append(Cell("%s - %s" % (tap_link, tap_desc), "lesson"))
                lesson_table.add_row(row)
            ofile.write(lesson_table.html())
            self.write_prev_next(ofile, prev, next, up)
            self.write_footer(ofile)

    def write_lesson_index(self):
        ofile, cwd = self.open_file(FileInfo.lesson_index)
        self.write_header(ofile, "Lessons", cwd, "lessons")
        li_table = Table("lesson", 'ID', 'Lesson')
        for i in range(self.skool.min_lesson, self.skool.max_lesson + 1):
            row = []
            lesson_file = FileInfo.lesson_relpath(cwd, i)
            row.append(Cell('<a class="link" href="%s">%i</a>' % (lesson_file, i), "lessonType"))
            row.append(Cell(self.skool.lesson_descs[i], "lesson"))
            li_table.add_row(row)
        ofile.write(li_table.html())
        self.write_footer(ofile)

    def write_personal_timetable_index(self):
        ofile, cwd = self.open_file(FileInfo.personal_timetables)
        self.write_header(ofile, "Personal timetables", cwd, "timetables")
        pt_table = Table("timetable", 'No.', 'Character')
        for c in range(self.skool.min_character, self.skool.max_character + 1):
            row = []
            row.append(Cell(self.expand('#PT%i' % c, cwd)))
            row.append(Cell(self.skool.characters[c]))
            pt_table.add_row(row)
        ofile.write(pt_table.html())
        self.write_footer(ofile)

    def write_memory_map(self):
        self.write_map(FileInfo.memory_map, 'Memory map', 'rdtgu', FileInfo.asm_path, True)

    def write_routines_map(self):
        self.write_map(FileInfo.routines, 'Routines', 'r', FileInfo.asm_path)

    def write_data_map(self):
        self.write_map(FileInfo.data, 'Data', 'd', FileInfo.asm_path, True)

    def write_messages_map(self):
        self.write_map(FileInfo.messages, 'Messages', 't', FileInfo.asm_path)

    def _write_characters_table(self, ofile, cwd):
        cols = 3

        headers = []
        for c in range(cols):
            headers.append(Header('Page'))
            headers.append(Header('Address'))
            headers.append(Header('Character'))
            if c < cols - 1:
                headers.append(Cell('', "transparent"))
        char_table = Table('cbuffer')
        char_table.add_row(headers)

        characters = self.skool.characters
        rows = (cols - 1  + len(characters)) // cols
        for r in range(rows):
            row = []
            for c in range(cols):
                char_index = r + c * rows
                if char_index < len(characters):
                    char_num = self.skool.min_character + char_index
                    row.append(Cell(char_num, "centre"))
                    buf_addr = self.skool.char_buffer_base_byte + char_num * 256
                    entry_link = self.expand('#R%i' % buf_addr, cwd)
                    row.append(Cell(entry_link, "centre"))
                    row.append(Cell(characters[char_num]))
                    if c < cols - 1:
                        row.append(Cell('', "transparent"))
            char_table.add_row(row)
        ofile.write(char_table.html())

    def write_character_buffers(self):
        ofile, cwd = self.open_file(FileInfo.cbuffer)
        self.write_header(ofile, 'Character buffers', cwd, 'cbuffer')

        ofile.write('<p>%s</p>' % self.skool.game_vars['CharBufDesc'])

        self._write_characters_table(ofile, cwd)

        ofile.write('<p>The bytes in each character buffer are used as follows:</p>')

        cb_table = Table('cbuffer', 'Byte(s)', 'Contents')
        for bytes, entries in self.skool.get_ref_items('CharBuf'):
            rows = len(entries)
            anchors = ''.join(['<a name="%s" />' % name for name in bytes.split(',')])
            row = []
            row.append(Cell(bytes, "centre", rowspan=rows))
            row.append(Cell("%s%s" % (anchors, self.expand(cgi.escape(entries[0]), cwd)), "byteDesc"))
            cb_table.add_row(row)
            for i in range(1, rows):
                cb_table.add_row((Cell(self.expand(cgi.escape(entries[i]), cwd), "byteDesc"),))
        ofile.write(cb_table.html())

        self.write_footer(ofile)

    def write_gbuffer(self):
        ofile, cwd = self.open_file(FileInfo.gbuffer)
        self.write_header(ofile, "Game status buffer", cwd, "gbuffer")
        gb_table = Table("gbuffer", 'Address', 'Length', 'Purpose')
        for entry in self.skool.memory_map:
            address = int(entry.address)
            if not (32512 <= address <= 32767):
                continue
            desc = '    <div class="gbufDesc">%s</div>' % entry.description
            if entry.details:
                desc += '\n    <div class="gbufDetails">%s</div>' % entry.details
            row = []
            asm_file = FileInfo.asm_relpath(cwd, address)
            entry_link = '<a name="%i" class="link" href="%s">%i</a>' % (address, asm_file, address)
            row.append(Cell(entry_link, "bufAddress"))
            row.append(Cell(entry.size, "bufLength"))
            row.append(Cell("\n%s\n" % self.expand(desc, cwd), "bufDesc"))
            gb_table.add_row(row)
        ofile.write(gb_table.html())
        self.write_footer(ofile)

    def write_load_entries(self):
        self.write_entries(FileInfo.load_path, FileInfo.load_index, "Load routines")

    def write_load_map(self):
        self.write_map(FileInfo.load_index, 'Loading code', 'r', FileInfo.load_path)

    def write_save_routine(self):
        fname = FileInfo.save
        cwd = os.path.dirname(fname)
        save = os.path.basename(fname)
        save_routine = self.skool.entries.get(str(self.skool.save_routine_address))
        self.write_entry(cwd, save_routine, save)

    def write_glossary(self):
        ofile, cwd = self.open_file(FileInfo.glossary)
        self.write_header(ofile, 'Glossary', cwd, 'glossary')
        table = Table('glossary', 'Term', 'Description')
        for term, paragraphs in self.skool.get_ref_items(GLOSSARY):
            row = []
            row.append(Cell(term, "glossaryTerm"))
            row.append(Cell(self.expand(self.expand_vars(paragraphs[0]), cwd), "glossaryDefn"))
            table.add_row(row)
        ofile.write(table.html())
        self.write_footer(ofile)

    def write_link_list(self, ofile, link_list):
        ofile.write('<ul class="linkList">\n')
        for anchor, title in link_list:
            ofile.write('<li><a class="link" href="#%s">%s</a></li>\n' % (anchor, title))
        ofile.write('</ul>\n')

    def write_boxes(self, ofile, boxes, cwd):
        box_num = 1
        for anchor, title, paragraphs in boxes:
            ofile.write('<div class="box box%i">\n' % box_num)
            ofile.write('<a name="%s" />\n' % anchor)
            ofile.write('<div class="boxTitle">%s</div>\n' % title)
            for paragraph in paragraphs:
                ofile.write('<div class="paragraph">%s</div>\n' % self.expand(self.expand_vars(paragraph), cwd))
            ofile.write('</div>\n')
            box_num = 3 - box_num

    def write_box_page(self, fname, title, body_class, boxes):
        ofile, cwd = self.open_file(fname)
        self.write_header(ofile, title, cwd, body_class)
        self.write_link_list(ofile, [(anchor, title) for anchor, title, p in boxes])
        self.write_boxes(ofile, boxes, cwd)
        self.write_footer(ofile)

    def write_pokes(self):
        self.write_box_page(FileInfo.pokes, 'Pokes', 'pokes', self.skool.get_ref_items(POKE))

    def write_bugs(self):
        self.write_box_page(FileInfo.bugs, 'Bugs', 'bugs', self.skool.get_ref_items(BUG))

    def write_facts(self):
        self.write_box_page(FileInfo.facts, 'Trivia', 'facts', self.skool.get_ref_items(FACT))

    def write_prev_next(self, ofile, prev, next, up):
        if prev or next or up:
            prev_text = next_text = up_text = ''
            if prev:
                prev_text = 'Prev: %s' % prev
            if next:
                next_text = 'Next: %s' % next
            if up:
                up_text = 'Up: %s' % up
            pn_table = Table("prevNext")
            pn_table.add_row((Cell(prev_text, "prev"), Cell(up_text, "up"), Cell(next_text, "next")))
            ofile.write(pn_table.html())

    def write_entry(self, cwd, entry, fname=None, page_header=None, prev=None, next=None, up=None):
        fname = fname or FileInfo.asm_fname(entry.address)
        ofile = self.file_info.open_file(cwd, fname)
        body_class = "disassembly"
        if entry.is_routine():
            title_format = "Routine at %s"
            page_header = page_header or "Routines"
        elif entry.ctl in 'uz':
            title_format = "Unused RAM at %s"
            page_header = page_header or "Unused"
        elif entry.ctl == 'g':
            title_format = "Game status buffer entry at %s"
            page_header = page_header or "Game status buffer"
        else:
            title_format = "Data at %s"
            page_header = page_header or "Data"
        self.write_header(ofile, title_format % entry.address, cwd, body_class, page_header)

        if prev:
            prev_file = FileInfo.asm_fname(prev.address)
            prev = '<a class="link" href="%s">%s</a>' % (prev_file, prev.address)
        if next:
            next_file = FileInfo.asm_fname(next.address)
            next = '<a class="link" href="%s">%s</a>' % (next_file, next.address)
        if up:
            map_path = FileInfo.relpath(cwd, up)
            up = '<a class="link" href="%s#%s">Map</a>' % (map_path, entry.address)
        self.write_prev_next(ofile, prev, next, up)

        desc = self.expand(entry.description, cwd)
        ofile.write('<div class="description">%s: %s</div>\n' % (entry.address, desc))

        table_class = 'disassembly'
        comment_class = 'comment'
        transparent_class = 'transparentComment'
        if entry.ctl not in 'cuz':
            table_class = 'dataDisassembly'
            comment_class = 'dataComment'
            transparent_class = 'transparentDataComment'
        ofile.write('<table class="%s">\n' % table_class)
        details = self.expand(entry.details, cwd).strip()
        if details or entry.registers:
            ofile.write('<tr>\n')
            ofile.write('<td class="routineComment" colspan="3">\n')
            if details:
                ofile.write('<div class="details">\n%s\n</div>\n' % details)
            if entry.registers:
                ofile.write('<table class="registers">\n')
                for reg in entry.registers:
                    ofile.write('<tr>\n')
                    ofile.write('<td class="register">%s</td>\n' % reg.name)
                    ofile.write('<td class="registerContents">%s</td>\n' % self.expand(reg.contents, cwd))
                    ofile.write('</tr>\n')
                ofile.write('</table>\n')
            ofile.write('</td>\n')
            ofile.write('</tr>\n')
        show_comment_col = False
        for instruction in entry.instructions:
            if instruction.comment and instruction.comment.text:
                show_comment_col = True
        for instruction in entry.instructions:
            mid_routine_comments = self.expand(entry.get_mid_routine_comment(instruction.label), cwd)
            if mid_routine_comments:
                ofile.write('<tr>\n')
                ofile.write('<td class="routineComment" colspan="3">\n')
                ofile.write('<div class="comments">%s</div>\n' % mid_routine_comments)
                ofile.write('</td>\n')
                ofile.write('</tr>\n')
            ofile.write('<tr>\n')
            tdclass = 'label' if instruction.ctl in 'c*!' else 'address'
            address = instruction.address
            ofile.write('<td class="%s"><a name="%s">%s</a></td>\n' % (tdclass, address, address))
            operation, reference = instruction.operation, instruction.reference
            if reference:
                routine_address = reference.routine_address
                name = '' if reference.address == routine_address else '#%s' % reference.address
                routine_file = FileInfo.asm_fname(routine_address)
                link = '<a class="link" href="%s%s">%s</a>' % (routine_file, name, reference.address)
                operation = operation.replace(reference.address, link)
            ofile.write('<td class="instruction">%s</td>\n' % operation)
            if show_comment_col:
                comment = instruction.comment
                if comment:
                    rowspan = '' if comment.rowspan < 2 else ' rowspan="%i"' % comment.rowspan
                    ofile.write('<td class="%s"%s>%s</td>\n' % (comment_class, rowspan, self.expand(comment.text, cwd)))
            else:
                ofile.write('<td class="%s" />\n' % transparent_class)
            ofile.write('</tr>\n')

        end_routine_comments = self.expand(entry.end_comment, cwd)
        if end_routine_comments:
            ofile.write('<tr>\n')
            ofile.write('<td class="routineComment" colspan="3">\n')
            ofile.write('<div class="comments">%s</div>\n' % end_routine_comments)
            ofile.write('</td>\n')
            ofile.write('</tr>\n')

        ofile.write('</table>\n')

        self.write_prev_next(ofile, prev, next, up)
        self.write_footer(ofile)

    def write_entries(self, cwd, map_file, page_header=None):
        prev = None
        memory_map = self.skool.memory_map
        for i, entry in enumerate(memory_map):
            next = memory_map[i + 1] if i + 1 < len(memory_map) else None
            self.write_entry(cwd, entry, page_header=page_header, prev=prev, next=next, up=map_file)
            prev = entry

    def write_map(self, map_file, title, entry_types, asm_path, show_page_byte=False):
        show_routines = 'r' in entry_types
        show_data = 'd' in entry_types
        show_text = 't' in entry_types
        show_gbuffer = 'g' in entry_types
        show_unused = 'u' in entry_types
        cwd = os.path.dirname(map_file)

        ofile = self.file_info.open_file(map_file)
        self.write_header(ofile, title, cwd, "map")
        details = self.skool.map_details.get(entry_types)
        if details:
            ofile.write('<div class="details">%s</div>\n' % self.expand(details, cwd))
        headers = ['Address', 'Description']
        if show_page_byte:
            headers.insert(0, 'Page')
            headers.insert(1, 'Byte')
        table = Table("map", *headers)

        for entry in self.skool.memory_map:
            purpose = entry.description
            if entry.ctl == 'c' and show_routines:
                entry_class = 'routine'
                desc_class = 'purpose'
            elif entry.ctl in 'bw' and show_data:
                entry_class = 'data'
                desc_class = 'dataDesc'
            elif entry.ctl in 'g' and show_gbuffer:
                entry_class = 'gbuffer'
                desc_class = 'gbufferDesc'
            elif entry.ctl in 'uz' and show_unused:
                size = entry.size
                purpose = 'Unused (%s byte%s)' % (size, 's' if size > 1 else '')
                entry_class = 'unused'
                desc_class = 'unusedDesc'
            elif entry.ctl == 't' and show_text:
                entry_class = 'messageAddress'
                desc_class = 'message'
            else:
                continue

            address = entry.address
            row = []
            if show_page_byte:
                row.append(Cell(int(address) // 256, "centre"))
                row.append(Cell(int(address) % 256, "centre"))
            asm_relpath = FileInfo.relpath(cwd, asm_path)
            asm_file = FileInfo.asm_fname(address, asm_relpath)
            content = '<a class="link" name="%s" href="%s">%s</a>' % (address, asm_file, address)
            row.append(Cell(content, entry_class))
            row.append(Cell(self.expand(purpose, cwd), desc_class))
            table.add_row(row)

        ofile.write(table.html())
        self.write_footer(ofile)

    def open_file(self, fname):
        ofile = self.file_info.open_file(fname)
        cwd = os.path.dirname(fname)
        return ofile, cwd

    def write_prologue(self, ofile):
        ofile.write('<?xml version="1.0" encoding="utf-8" ?>\n')
        ofile.write('<!DOCTYPE html\n')
        ofile.write('    PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"\n')
        ofile.write('    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n')
        ofile.write('<html xmlns="http://www.w3.org/1999/xhtml">\n')
        ofile.write('<!-- Generated by skool2html on %s -->\n' % self.time)

    def write_head(self, ofile, title, cwd, js=False):
        self.write_prologue(ofile)
        ofile.write('<head>\n')
        ofile.write('<title>%s: %s</title>\n' % (self.skool.game, title))
        ofile.write('<link rel="stylesheet" type="text/css" href="%s" />\n' % FileInfo.css_relpath(cwd))
        if js:
            ofile.write('<script type="text/javascript" src="%s"></script>\n' % FileInfo.js_relpath(cwd))
        ofile.write('</head>\n')

    def _get_logo(self, cwd):
        if self.file_info.file_exists(FileInfo.logo):
            return '<img src="%s" alt="%s" />' % (FileInfo.relpath(cwd, FileInfo.logo), self.skool.game)
        return self.skool.game

    def write_header(self, ofile, title, cwd, body_class, body_title=None, js=False):
        self.write_head(ofile, title, cwd, js)
        index = FileInfo.relpath(cwd, FileInfo.index)
        ofile.write('<body class="%s">\n' % body_class)
        table = Table("header")
        row = []
        row.append(Cell('<a class="link" href="%s">%s</a>' % (index, self._get_logo(cwd))))
        row.append(Cell(body_title or title, "headerText"))
        table.add_row(row)
        ofile.write(table.html())

    def write_footer(self, ofile):
        ofile.write('</body>\n')
        ofile.write('</html>\n')
        ofile.close()

    def write_logo_image(self):
        logo = self.skool.get_logo()
        if logo:
            self.write_image(logo, FileInfo.logo)

    def get_pixel_array(self, udg_array, mask):
        # Convert UDG array into byte array
        byte_array = []
        for row in udg_array:
            for line in range(8):
                byte_array.append([(udg.attr, udg.data[line], udg.mask[line] if mask else 0) for udg in row])

        # Convert byte array into pixel array
        pixel_array = []
        for row in byte_array:
            pixel_array.append([])
            for attr, byte, mask_byte in row:
                bright = (attr & 64) // 8
                ink = bright + (attr & 7)
                paper = bright + (attr & 56) // 8
                for b in range(8):
                    attr = paper
                    pixel = byte & 128
                    if mask:
                        if pixel & mask_byte:
                            attr = ink
                        elif mask_byte & 128:
                            attr = 16
                        mask_byte *= 2
                    elif pixel:
                        attr = ink
                    pixel_array[-1].append(attr)
                    byte *= 2

        return pixel_array

    def _gd_create_colours(self, image):
        colours = []
        colours.append(image.colorAllocate(BLACK))
        colours.append(image.colorAllocate(BLUE))
        colours.append(image.colorAllocate(RED))
        colours.append(image.colorAllocate(MAGENTA))
        colours.append(image.colorAllocate(GREEN))
        colours.append(image.colorAllocate(CYAN))
        colours.append(image.colorAllocate(YELLOW))
        colours.append(image.colorAllocate(WHITE))
        colours.append(colours[0])
        colours.append(image.colorAllocate(BRIGHT_BLUE))
        colours.append(image.colorAllocate(BRIGHT_RED))
        colours.append(image.colorAllocate(BRIGHT_MAGENTA))
        colours.append(image.colorAllocate(BRIGHT_GREEN))
        colours.append(image.colorAllocate(BRIGHT_CYAN))
        colours.append(image.colorAllocate(BRIGHT_YELLOW))
        colours.append(image.colorAllocate(BRIGHT_WHITE))
        colours.append(image.colorAllocate(TRANSPARENT))
        return colours

    def _gd_write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        pixel_array = self.get_pixel_array(udgs, mask)
        width = width or len(pixel_array[0])
        image = gd.image((scale * width, scale * len(pixel_array)))
        colours = self._gd_create_colours(image)
        for y, row in enumerate(pixel_array):
            for x, attr in enumerate(row[:width]):
                x1, y1 = x * scale, y * scale
                x2, y2 = x1 + scale - 1, y1 + scale - 1
                image.filledRectangle((x1, y1), (x2, y2), colours[attr])
        f = self.file_info.open_file(img_file)
        image.writePng(f)
        f.close()

    def _pil_create_colours(self):
        self.colours = []
        self.colours.extend(BLACK)
        self.colours.extend(BLUE)
        self.colours.extend(RED)
        self.colours.extend(MAGENTA)
        self.colours.extend(GREEN)
        self.colours.extend(CYAN)
        self.colours.extend(YELLOW)
        self.colours.extend(WHITE)
        self.colours.extend(BRIGHT_BLACK)
        self.colours.extend(BRIGHT_BLUE)
        self.colours.extend(BRIGHT_RED)
        self.colours.extend(BRIGHT_MAGENTA)
        self.colours.extend(BRIGHT_GREEN)
        self.colours.extend(BRIGHT_CYAN)
        self.colours.extend(BRIGHT_YELLOW)
        self.colours.extend(BRIGHT_WHITE)
        self.colours.extend(TRANSPARENT)

    def _pil_write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        pixel_array = self.get_pixel_array(udgs, mask)
        width = width or len(pixel_array[0])
        image = Image.new('P', (scale * width, scale * len(pixel_array)), None)
        image.putpalette(self.colours)
        pixels = []
        for row in pixel_array:
            pixel_row = []
            for pixel in row[:width]:
                pixel_row.extend([pixel] * scale)
            for i in range(scale):
                pixels.extend(pixel_row)
        image.putdata(pixels)
        f = self.file_info.open_file(img_file)
        image.save(f, 'PNG', optimize=True)
        f.close()

    write_image = _pil_write_image if USE_PIL else _gd_write_image

    def _create_macros(self):
        self.macros = {}
        self.macros['#AS'] = self.expand_as
        self.macros['#ASIMG'] = self.expand_asimg
        self.macros['#ASM'] = self.expand_asm
        self.macros['#ASTILE'] = self.expand_astile
        self.macros['#BOX'] = self.expand_box
        self.macros['#BUG'] = self.expand_bug
        self.macros['#C'] = self.expand_c
        self.macros['#CBUF'] = self.expand_cbuf
        self.macros['#D'] = self.expand_desc
        self.macros['#EREFS'] = self.expand_erefs
        self.macros['#FACT'] = self.expand_fact
        self.macros['#FONT'] = self.expand_font
        self.macros['#GBUF'] = self.expand_gbuf
        self.macros['#KEYTABLE'] = self.expand_keytable
        self.macros['#LESSON'] = self.expand_lesson
        self.macros['#LOAD'] = self.expand_load
        self.macros['#MSG'] = self.expand_msg
        self.macros['#MUTABLE'] = self.expand_mutable
        self.macros['#NAME'] = self.expand_name
        self.macros['#P'] = self.expand_p
        self.macros['#PATILE'] = self.expand_patile
        self.macros['#POKE'] = self.expand_poke
        self.macros['#POKES'] = self.expand_pokes
        self.macros['#POPS'] = self.expand_pops
        self.macros['#PT'] = self.expand_pt
        self.macros['#PUSHS'] = self.expand_pushs
        self.macros['#R'] = self.expand_reference
        self.macros['#REFS'] = self.expand_refs
        self.macros['#REG'] = self.expand_reg
        self.macros['#SAVE'] = self.expand_save
        self.macros['#SCR'] = self.expand_scr
        self.macros['#SKOOL'] = self.expand_skool
        self.macros['#SPACE'] = self.expand_space
        self.macros['#START'] = self.expand_start
        self.macros['#TABLE'] = self.expand_table
        self.macros['#TAP'] = self.expand_tap
        self.macros['#TAPS'] = self.expand_taps
        self.macros['#TIME'] = None
        self.macros['#UDG'] = self.expand_udg
        self.macros['#UDGTABLE'] = self.expand_table

    def error(self, msg):
        sys.stderr.write('\n*** Error: %s\n' % msg)
        sys.exit(1)

    def build_table(self, table):
        table_class = table.table_class
        html = '\n<table%s>\n' % (' class="%s"' % table_class if table_class else '')
        for row in table.rows:
            html += '<tr>\n'
            for cell in row:
                colspan = ' colspan="%i"' % cell.colspan if cell.colspan > 1 else ''
                rowspan = ' rowspan="%i"' % cell.rowspan if cell.rowspan > 1 else ''
                if cell.header:
                    tag = 'th'
                    td_class = ''
                else:
                    tag = 'td'
                    cell_class = cell.cell_class
                    if cell.transparent:
                        cell_class += " transparent"
                    cell_class = cell_class.lstrip()
                    td_class = ' class="%s"' % cell_class if cell_class else ''
                html += '<%s%s%s%s>%s</%s>\n' % (tag, td_class, colspan, rowspan, cell.contents, tag)
            html += '</tr>\n'
        return html + '</table>\n'

    def get_int_params(self, text, index, num, req, *defaults):
        param_string = ''
        commas = 0
        while index < len(text) and text[index] in '0123456789,':
            char = text[index]
            if char == ',':
                commas += 1
                if commas == num:
                    break
            param_string += char
            index += 1
        params = [int(p) for p in param_string.split(',')]
        i = len(params) - req
        while len(params) < num:
            params.append(defaults[i] if i < len(defaults) else None)
            i += 1
        return [index] + params

    def get_str_params(self, text, index, link_text='', chars=''):
        start = index
        valid_chars = '#' + chars
        while index < len(text) and (text[index] in valid_chars or text[index].isalnum()):
            index += 1
        param = text[start:index]
        end = index
        if index < len(text) and text[index] == '(':
            end = text.index(')', index) + 1
            link_text = text[index + 1:end - 1]
        return end, param, link_text

    def expand_as(self, text, index, cwd):
        # #AS[state][(link text)]
        end, state, link_text = self.get_str_params(text, index)
        link_text = link_text or state
        anchor = '#%s' % state if state else ''
        as_file = FileInfo.relpath(cwd, FileInfo.animatory_states)
        link = '<a class="link" href="%s%s">%s</a>' % (as_file, anchor, link_text)
        return end, link

    def expand_asimg(self, text, index, cwd):
        # #ASIMGnum[,scale[,mask[,attribute]]]
        end, num, scale, mask, attr = self.get_int_params(text, index, 4, 1, 2, 1, 120)
        asimg_fname = FileInfo.as_fname(num, attr, scale, mask, self.skool.get_snapshot_name())
        if not self.file_exists(asimg_fname):
            self.write_image(self.skool.build_sprite(num, attr), asimg_fname, scale=scale, mask=mask)
        img_path = FileInfo.relpath(cwd, asimg_fname)
        return end, '<img alt="Animatory state %i" src="%s" />' % (num, img_path)

    def expand_asm(self, text, index, cwd):
        # #ASMaddress[#anchor][(link text)]
        end = index + 5
        address = text[index:end]
        end, anchor, link_text = self.get_str_params(text, end, address)
        ref_file = FileInfo.asm_relpath(cwd, address)
        link = '<a class="link" href="%s%s">%s</a>' % (ref_file, anchor, link_text)
        return end, link

    def expand_astile(self, text, index, cwd, tile=None):
        # #ASTILEstate,row,col[,attr[,mask[,scale]]]
        end, state, row, col, attr, mask, scale = self.get_int_params(text, index, 6, 3, 120, 1, 4)
        if tile is None:
            tile = self.skool.get_sprite_udg(state, attr, row, col)
        img_fname = FileInfo.astile_fname(tile.udg_page, tile.ref, tile.attr, scale, mask)
        if not self.file_exists(img_fname):
            self.write_image([[tile]], img_fname, scale, mask)
        img_path = FileInfo.relpath(cwd, img_fname)
        return end, '<img alt="%i:%i,%i" src="%s"/>' % (state, row, col, img_path)

    def expand_box(self, text, index, cwd):
        # #BOXaddress[,scale[,attribute]]
        end, address, scale, attr = self.get_int_params(text, index, 3, 1, 2)
        box_fname = FileInfo.box_fname(address, attr, scale, self.skool.get_snapshot_name())
        if not self.file_exists(box_fname):
            self.write_image(self.skool.get_box_graphic(address, attr), box_fname, scale)
        img_path = FileInfo.relpath(cwd, box_fname)
        return end, '<img alt="BOX%s" src="%s" />' % (address, img_path)

    def expand_bug(self, text, index, cwd):
        # #BUG[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, 'bug')
        bugs_file = FileInfo.relpath(cwd, FileInfo.bugs)
        link = '<a class="link" href="%s%s">%s</a>' % (bugs_file, anchor, link_text)
        return end, link

    def expand_c(self, text, index, cwd):
        # #C
        return index, '&#169;'

    def expand_cbuf(self, text, index, cwd):
        # #CBUF[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index)
        link_text = link_text or anchor[1:] or 'buffer'
        cbuffer_file = FileInfo.relpath(cwd, FileInfo.cbuffer)
        link = '<a class="link" href="%s%s">%s</a>' % (cbuffer_file, anchor, link_text)
        return end, link

    def expand_desc(self, text, index, cwd):
        # #Daddress
        end, address = self.get_int_params(text, index, 1, 1)
        entry = self.skool.get_entry(address)
        desc = address
        if entry:
            if entry.description:
                desc = entry.description
            else:
                self.error('Entry at %s has no description' % address)
        else:
            self.error('Cannot determine description for nonexistent entry at %s' % address)
        return end, str(desc)

    def expand_erefs(self, text, index, cwd):
        # #EREFSaddress
        end, address = self.get_int_params(text, index, 1, 1)
        ereferrers = self.skool.get_entry_point_refs(address)
        if not ereferrers:
            self.error('Entry point at %i has no referrers' % address)
        ereferrers.sort()
        eref_files = [(FileInfo.asm_relpath(cwd, addr), addr) for addr in referrers]
        html = 'routine at '
        if len(ereferrers) > 1:
            html = 'routines at '
            html += ', '.join('<a class="link" href="%s">%s</a>' % (eref_file, addr) for eref_file, addr in eref_files[:-1])
            html += ' and '
        eref_file, addr = eref_files[-1]
        html += '<a class="link" href="%s">%s</a>' % (eref_file, addr)
        return end, html

    def expand_fact(self, text, index, cwd):
        # #FACT[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, 'fact')
        facts_file = FileInfo.relpath(cwd, FileInfo.facts)
        link = '<a class="link" href="%s%s">%s</a>' % (facts_file, anchor, link_text)
        return end, link

    def expand_font(self, text, index, cwd):
        # #FONT
        font_fname = FileInfo.font
        if not self.file_exists(font_fname):
            char_widths = self.skool.snapshot[self.skool.font_address + 32:self.skool.font_address + 128]
            width = 1 + sum(char_widths) + len(char_widths)
            udgs = self.skool.get_text(''.join([chr(c) for c in range(32, 128)]))
            font_udg_array = [[Udg(56, udg) for udg in udgs]]
            self.write_image(font_udg_array, font_fname, width=width)
        img_path = FileInfo.relpath(cwd, font_fname)
        return index, '<img alt="font" src="%s" />' % img_path

    def expand_gbuf(self, text, index, cwd):
        # #GBUF[address][(link text)]
        end, address, link_text = self.get_str_params(text, index)
        gbuffer_file = FileInfo.relpath(cwd, FileInfo.gbuffer)
        anchor = '#%s' % address if address else ''
        link = '<a class="link" href="%s%s">%s</a>' % (gbuffer_file, anchor, link_text or address or 'game status buffer')
        return end, link

    def expand_keytable(self, text, index, cwd):
        # #KEYTABLE[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, 'keypress table')
        keys_file = FileInfo.relpath(cwd, FileInfo.keys)
        link = '<a class="link" href="%s%s">%s</a>' % (keys_file, anchor, link_text)
        return end, link

    def expand_lesson(self, text, index, cwd):
        # #LESSONnum
        end, lesson_num = self.get_int_params(text, index, 1, 1)
        lesson_file = FileInfo.lesson_relpath(cwd, lesson_num)
        link = '<a class="link" href="%s">%i</a>' % (lesson_file, lesson_num)
        return end, link

    def expand_load(self, text, index, cwd):
        # #LOADaddress[#anchor][(link text)]
        end = index + 5
        address = text[index:end]
        end, anchor, link_text = self.get_str_params(text, end, address)
        load_file = FileInfo.load_relpath(cwd, address)
        link = '<a class="link" href="%s%s">%s</a>' % (load_file, anchor, link_text)
        return end, link

    def expand_msg(self, text, index, cwd):
        # #MSGnum[(link text)], #MSGaddress[(link text)]
        end, msg_num, link_text = self.get_str_params(text, index)
        link_text = link_text or msg_num
        address = self.skool.message_addresses.get(msg_num, msg_num)
        if address not in self.skool.entries:
            self.error('No entry for message at %s' % address)
        asm_file = FileInfo.asm_relpath(cwd, address)
        link = '<a class="link" href="%s">%s</a>' % (asm_file, link_text)
        return end, link

    def expand_mutable(self, text, index, cwd):
        # #MUTABLEaddress[,padding]
        end, address, padding = self.get_int_params(text, index, 2, 1, 0)
        mutable_fname = FileInfo.mutable_fname(address, padding)
        if not self.file_exists(mutable_fname):
            mutable_udgs = self.skool.get_mutable_udg_array(address, padding)
            self.write_image(mutable_udgs, mutable_fname, 2)
        img_path = FileInfo.relpath(cwd, mutable_fname)
        return end, '<img alt="MUTABLE%s" src="%s" />' % (address, img_path)

    def expand_name(self, text, index, cwd):
        # #NAMEname
        end, name, link_text = self.get_str_params(text, index)
        return end, '<a name="%s" />' % name

    def expand_p(self, text, index, cwd):
        # #P
        return index, '<br /><br />'

    def expand_patile(self, text, index, cwd, tile=None):
        # #PATILEx,y[,scale]
        end, x, y, scale = self.get_int_params(text, index, 3, 2, 4)
        if tile is None:
            tile = self.skool.get_skool_udg(y, x)
        img_fname = FileInfo.patile_fname(tile.udg_page, tile.ref, tile.attr, scale)
        if not self.file_exists(img_fname):
            self.write_image([[tile]], img_fname, scale)
        img_path = FileInfo.relpath(cwd, img_fname)
        return end, '<img alt="%i,%i" src="%s"/>' % (x, y, img_path)

    def expand_poke(self, text, index, cwd):
        # #POKE[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, 'poke')
        pokes_file = FileInfo.relpath(cwd, FileInfo.pokes)
        link = '<a class="link" href="%s%s">%s</a>' % (pokes_file, anchor, link_text)
        return end, link

    def expand_pokes(self, text, index, cwd):
        # #POKESaddress,byte[,length[,step]]
        end, address, byte, length, step = self.get_int_params(text, index, 4, 2, 1, 1)
        self.skool.snapshot[address:address + length * step:step] = [byte] * length
        return end, ''

    def expand_pops(self, text, index, cwd):
        # #POPS
        self.skool.pop_snapshot()
        return index, ''

    def expand_pt(self, text, index, cwd):
        # #PTnum[(link text)]
        index, char_num = self.get_int_params(text, index, 1, 1)
        end, x, link_text = self.get_str_params(text, index, str(char_num))
        pt_address = self.skool.min_lesson + 256 * char_num
        pt_file = FileInfo.asm_relpath(cwd, pt_address)
        link = '<a class="link" href="%s">%s</a>' % (pt_file, link_text)
        return end, link

    def expand_pushs(self, text, index, cwd):
        # #PUSHSname
        end, name, link_text = self.get_str_params(text, index)
        self.skool.push_snapshot(name)
        return end, ''

    def expand_refs(self, text, index, cwd):
        # #REFSaddress[(prefix)]
        end, address = self.get_int_params(text, index, 1, 1)
        end, address, prefix = self.get_str_params(text, index, '')
        entry = self.skool.entries.get(address)
        if not entry:
            self.error('Cannot determine referrers for nonexistent entry at %i' % address)
        referrers = [ref.address for ref in entry.referrers]
        if referrers:
            referrers.sort()
            ref_files = [(FileInfo.asm_relpath(cwd, addr), addr) for addr in referrers]
            html = ('%s routine at ' % prefix).lstrip()
            if len(referrers) > 1:
                html = ('%s routines at ' % prefix).lstrip()
                html += ', '.join('<a class="link" href="%s">%s</a>' % (ref_file, addr) for ref_file, addr in ref_files[:-1])
                html += ' and '
            ref_file, addr = ref_files[-1]
            html += '<a class="link" href="%s">%s</a>' % (ref_file, addr)
        else:
            html = 'Not used directly by any other routines'
        return end, html

    def expand_reference(self, text, index, cwd):
        # #Raddress[#anchor][(link text)]
        end = index + 5
        address = text[index:end]
        container = self.skool.get_container(address)
        if not container:
            self.error('Could not find routine file containing %s' % address)
        end, anchor, link_text = self.get_str_params(text, end, address)
        anchor = anchor if address == container.address else '#%s' % address
        ref_file = FileInfo.asm_relpath(cwd, container.address)
        link = '<a class="link" href="%s%s">%s</a>' % (ref_file, anchor, link_text)
        return end, link

    def expand_reg(self, text, index, cwd):
        # #REGx[y][']
        end = index
        while end < len(text) and text[end] in "abcdefhlirspxy'":
            end += 1
        reg = text[index:end]
        if not reg:
            self.error('#REG requires a register argument')
        elif len(reg) > 3:
            self.error('Bad register: "#REG%s"' % reg)
        return end, '<span class="register">%s</span>' % reg.upper()

    def expand_save(self, text, index, cwd):
        # #SAVE[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, 'save routine')
        save_file = FileInfo.relpath(cwd, FileInfo.save)
        link = '<a class="link" href="%s%s">%s</a>' % (save_file, anchor, link_text)
        return end, link

    def expand_scr(self, text, index, cwd):
        # #SCR
        scr_fname = FileInfo.scr_fname(self.skool.skoolfile)
        if not self.file_exists(scr_fname):
            self.write_image(self.skool.get_screenshot(), scr_fname, 1)
        img_path = FileInfo.relpath(cwd, scr_fname)
        return index, '<img alt="screenshot" src="%s" />' % img_path

    def expand_skool(self, text, index, cwd):
        # #SKOOLx,y,w,h[,scale[,showChars[,showX]]]
        end, x, y, w, h, scale, show_chars, show_x = self.get_int_params(text, index, 7, 4, 2, 0, 0)
        skoolimg_fname = FileInfo.skoolimg_fname(x, y, w, h, scale, show_chars, show_x, self.skool.get_snapshot_name())
        if not self.file_exists(skoolimg_fname):
            self.write_image(self.skool.get_skool_udgs(x, y, w, h, show_chars, show_x), skoolimg_fname, scale=scale)
        img_path = FileInfo.relpath(cwd, skoolimg_fname)
        return end, '<img alt="Skool" src="%s" />' % img_path

    def expand_space(self, text, index, cwd):
        # #SPACE[num]
        end, num_sp = self.get_int_params(text, index, 1, 0, 1)
        return end, '&nbsp;' * num_sp

    def expand_start(self, text, index, cwd):
        # #START[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, str(self.skool.start_address))
        start_file = FileInfo.asm_relpath(cwd, self.skool.start_address)
        link = '<a class="link" href="%s%s">%s</a>' % (start_file, anchor, link_text)
        return end, link

    def expand_table(self, text, index, cwd):
        # #TABLE[(class[,col1class[,col2class...]])]<rows>TABLE#
        end, table = self.table_parser.parse_text(text, index)
        return end, self.build_table(table)

    def expand_taps(self, text, index, cwd):
        # #TAPSaddress
        end, address = self.get_int_params(text, index, 1, 1)
        referring_taps = self.skool.get_tap_refs(address)
        if not referring_taps:
            self.error('Routine at %i is not used by any command lists' % address)
        asm_files = [(FileInfo.asm_relpath(cwd, address), tap) for tap, address in referring_taps]
        html = 'command list '
        if len(referring_taps) > 1:
            html = 'command lists '
            html += ', '.join('<a class="link" href="%s">%s</a>' % (asm_file, tap) for asm_file, tap in asm_files[:-1])
            html += ' and '
        asm_file, tap = asm_files[-1]
        html += '<a class="link" href="%s">%s</a>' % (asm_file, tap)
        return end, html

    def expand_tap(self, text, index, cwd):
        # #TAPnum
        end, tap_num = self.get_int_params(text, index, 1, 1)
        tap_address = self.skool.tap_addresses[tap_num]
        asm_file = FileInfo.asm_relpath(cwd, tap_address)
        link = '<a class="link" href="%s">%s</a>' % (asm_file, tap_num)
        return end, link

    def expand_udg(self, text, index, cwd):
        # #UDGaddress,attribute[,scale[,step[,increment]]]
        end, address, attr, scale, step, inc = self.get_int_params(text, index, 5, 2, 4, 1, 0)
        udg_fname = FileInfo.udg_fname(address, attr, scale)
        if not self.file_exists(udg_fname):
            udg_bytes = [self.skool.snapshot[address + n * step] + inc for n in range(8)]
            self.write_image([[Udg(attr, udg_bytes)]], udg_fname, scale)
        img_path = FileInfo.relpath(cwd, udg_fname)
        return end, '<img alt="UDG%s" src="%s" />' % (address, img_path)

    def expand(self, text, cwd, **kwargs):
        if not text or text.find('#') < 0:
            return text

        while True:
            search = re.search('#[A-Z]+', text)
            if not search:
                break
            marker = search.group()
            if not marker in self.macros:
                self.error('Found unknown macro: %s' % marker)
            repf = self.macros[marker]
            if repf is None:
                self.error('Found unsupported macro: %s' % marker)
            start, index = search.span()
            end, rep = repf(text, index, cwd, **kwargs)
            text = text.replace(text[start:end], rep, 1)

        return text

    def expand_vars(self, text):
        if not text or text.find('${') < 0:
            return text

        while True:
            search = re.search('\${[a-z]+:[^}]+}', text)
            if not search:
                break
            macro = search.group()
            macro_id, value = macro[2:-1].split(':', 1)
            if macro_id == 'f':
                path_elements = []
                for p in value.split('/'):
                    if p.find('(') < 0:
                        path_elements.append(getattr(FileInfo, p))
                    else:
                        index = p.index('(')
                        method = getattr(FileInfo, p[:index])
                        params = p[index + 1:p.index(')')].split(',')
                        path_elements.append(method(*params))
                rep = join(*path_elements)
            elif macro_id == 'v':
                rep = self.skool.game_vars[value]
            else:
                rep = value if self.skool.game_id == macro_id else ''
            text = text.replace(macro, rep, 1)

        return text

class SkoolDazeWriter(SkoolWriter):
    def write_graphics(self):
        ofile, cwd = SkoolWriter.write_graphics(self)

        ofile.write('<h3>Score box, lines bubble, speech bubble</h3>\n')
        graphics = []
        graphics.append(('#BOX60928', 'Score box'))
        graphics.append(('#BOX60160,2,23', 'Lines bubble template'))
        scale = 2
        bubble_fname = FileInfo.bubble_img
        bubble_attr = 56
        if not self.file_exists(bubble_fname):
            bubble_udgs = [[], []]
            for n in range(8):
                bubble_udgs[0].append(Udg(bubble_attr, self.skool.snapshot[33008 + n:35056 + n:256]))
                bubble_udgs[1].append(Udg(bubble_attr, self.skool.snapshot[33016 + n:35064 + n:256]))
            self.write_image(bubble_udgs, bubble_fname, scale)
        img_path = FileInfo.relpath(cwd, bubble_fname)
        graphics.append(('<img alt="bubble" src="%s" />' % img_path, 'Speech bubble'))
        graphics.append(('#UDG33007,%i,%i,256' % (bubble_attr, scale), 'Speech bubble lip'))

        table = Table()
        table.cellpadding = 5
        for macro, desc in graphics:
            table.add_row((Cell(self.expand(macro, cwd)), Cell(desc)))
        ofile.write(table.html())

        ofile.write('<h3>Font</h3>\n')
        ofile.write(self.expand('<div>#FONT</div>\n', cwd))

        self.write_footer(ofile)

    def write_keytable(self):
        ofile, cwd, keypress_routines = SkoolWriter.write_keytable(self)
        snapshot = self.skool.snapshot

        table = Table('keytable', 'Code', 'Key', 'Address', 'Offset', 'Lookup', 'Keycode', 'Routine', 'Purpose')
        for c in range(32, 128):
            key = '&#169;' if c == 127 else cgi.escape(chr(c))
            address = c + 26624
            offset = snapshot[address]
            lookup = offset + 26624
            key_byte = snapshot[lookup]
            keycode = 'SPACE' if key_byte == 32 else chr(key_byte)
            routine = snapshot[lookup + 1] + 256 * snapshot[lookup + 2]
            purpose = keypress_routines[routine]
            routine_link = self.expand('#R%i' % routine, cwd)
            row = []
            row.append(Cell(c, "keyIndex"))
            row.append(Cell(key, "key"))
            row.append(Cell(address, "keyAddress"))
            row.append(Cell(offset, "data"))
            row.append(Cell(lookup, "keyAddress"))
            row.append(Cell(keycode, "centre"))
            row.append(Cell(routine_link, "keyAddress"))
            row.append(Cell(purpose, "purpose"))
            table.add_row(row)
        ofile.write(table.html())

        self.write_footer(ofile)

    def get_playarea_tile_text(self, tile):
        text = ['<i>ref_addr</i>=(x,y+152)=%i' % tile.ref_addr]
        text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
        text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
        text.append('<i>attr_addr</i>=(x+128,y+152)=%i' % tile.attr_addr)
        text.append('<i>attr</i>=%i' % tile.attr)
        return text

class BackToSkoolWriter(SkoolWriter):
    def write_mutable_tile_tables(self, ofile, cwd):
        doors, windows, bike = self.skool.get_mutables()

        sections = (('Doors', doors), ('Windows (open)', windows), ('Tree (without bike)', bike))
        for title, section in sections:
            ofile.write('<h3>%s</h3>\n<table class="mgrid">\n<tr>\n' % title)
            for a, initial_state, desc in section:
                addr = a + 256 if initial_state else a
                mutable_udgs = self.skool.get_mutable_udg_array(addr, -1)
                ofile.write('<td>\n')
                self.write_playarea_tile_table(ofile, cwd, mutable_udgs, 'm')
                ofile.write('</td>\n')
            ofile.write('</tr>\n</table>\n')

        cups = []
        cups.append((56694, "Leftmost cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((56950, "Middle cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((57206, "Rightmost cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((57462, "Cup on the shelf in the girls' skool (water/sherry)"))
        ofile.write('<h3>Cups</h3>\n<table>\n')
        for a, desc, in cups:
            ofile.write('<tr>\n<td>\n<table class="mgrid">\n<tr>\n')
            for address in (a, a + 5):
                cup_udg_array = self.skool.get_mutable_udg_array(address, -1)
                ofile.write('<td>\n')
                self.write_playarea_tile_table(ofile, cwd, cup_udg_array, 'c%i_' % address)
                ofile.write('</td>\n')
            ofile.write('</tr>\n</table>\n</td><td>%s</td></tr>\n' % desc)
        ofile.write('</table>\n')

    def write_graphics(self):
        ofile, cwd = SkoolWriter.write_graphics(self)

        doors, windows, bike = self.skool.get_mutables()
        sections = (('Doors', doors), ('Windows', windows), ('Bike', bike))
        for title, section in sections:
            ofile.write('<h3>%s</h3>' % title)
            table = Table()
            table.cellpadding = 5
            for a, initial_state, desc in section:
                row = []
                for address in ((a, a + 256)):
                    row.append(Cell(self.expand('#MUTABLE%i,1' % address, cwd)))
                row.append(Cell(desc))
                table.add_row(row)
            ofile.write(table.html())

        cups = []
        cups.append((56081, "All cups on the shelf in the boys' skool (empty)", []))
        cups.append((56694, "Leftmost cup on the shelf in the boys' skool (water/sherry)", [5]))
        cups.append((56950, "Middle cup on the shelf in the boys' skool (water/sherry)", [5]))
        cups.append((57206, "Rightmost cup on the shelf in the boys' skool (water/sherry)", [5]))
        cups.append((56337, "Cup on the shelf in the girls' skool (empty/water/sherry)", [1125, 5]))
        ofile.write('<h3>Cups</h3>\n')
        table = Table()
        table.cellpadding = 5
        for a, desc, offsets in cups:
            row = []
            subtable = Table()
            subrow = []
            address = a
            for offset in [0] + offsets:
                address += offset
                subrow.append(Cell(self.expand('#MUTABLE%i,1' % address, cwd)))
            subtable.add_row(subrow)
            row.append(Cell(subtable.html()))
            row.append(Cell(desc))
            table.add_row(row)
        ofile.write(table.html())

        scale = 2
        ofile.write('<h3>Inventory</h3>\n')
        table = Table()
        table.cellpadding = 10
        for addresses, attr, desc in self.skool.get_inventory_items():
            row = []
            if len(addresses) == 1:
                row.append(Cell(self.expand('#UDG%i,%i,%i' % (addresses[0], attr, scale), cwd)))
            else:
                basename = 'pistol'
                img_fname = FileInfo.udg_matrix_fname(basename, attr, scale)
                if not self.file_exists(img_fname):
                    udg_array = [[self.skool.get_udg(address, attr) for address in addresses]]
                    self.write_image(udg_array, img_fname, scale)
                img_path = FileInfo.relpath(cwd, img_fname)
                row.append(Cell('<img alt="%s" src="%s" />' % (basename, img_path)))
            row.append(Cell(desc))
            table.add_row(row)
        ofile.write(table.html())

        ofile.write('<h3>Score box and speech bubble</h3>\n')
        bubble_fname = FileInfo.bubble_img
        bubble_attr = 56
        if not self.file_exists(bubble_fname):
            top_row = [39392]
            top_row.extend([39648] * 6)
            top_row.append(39904)
            bottom_row = [40160]
            bottom_row.extend([40416] * 6)
            bottom_row.append(40672)
            bubble_udgs = [[self.skool.get_udg(address, bubble_attr) for address in top_row]]
            bubble_udgs.append([self.skool.get_udg(address, bubble_attr) for address in bottom_row])
            self.write_image(bubble_udgs, bubble_fname, scale)
        bubble_img_path = FileInfo.relpath(cwd, bubble_fname)
        others = []
        others.append(('#BOX58920', 'Score box'))
        others.append(('<img alt="bubble" src="%s" />' % bubble_img_path, 'Speech bubble'))
        others.append(('#UDG39136,%i,%i' % (bubble_attr, scale), 'Speech bubble lip'))
        table = Table()
        table.cellpadding = 5
        for contents, desc in others:
            table.add_row((Cell(self.expand(contents, cwd)), Cell(desc)))
        ofile.write(table.html())

        ofile.write('<h3>Font</h3>\n')
        ofile.write(self.expand('<div>#FONT</div>\n', cwd))

        self.write_footer(ofile)

    def write_keytable(self):
        ofile, cwd, keypress_routines = SkoolWriter.write_keytable(self)
        snapshot = self.skool.snapshot

        table = Table('keytable', 'Index', 'Key', 'Address', 'Offset', 'Lookup', 'Routine', 'Purpose')
        for index in range(80):
            key = '&#169;' if index == 79 else cgi.escape(chr(index + 48))
            address = index + 58624
            offset = snapshot[address]
            lookup = routine_link = purpose = ''
            if offset:
                lookup = offset + 58624
                routine = snapshot[lookup] + 256 * snapshot[lookup + 1]
                purpose = keypress_routines[routine]
                routine_link = self.expand('#R%i' % routine, cwd)
            row = []
            row.append(Cell(index, "keyIndex"))
            row.append(Cell(key, "key"))
            row.append(Cell(address, "keyAddress"))
            row.append(Cell(offset, "data"))
            row.append(Cell(lookup, "keyAddress"))
            row.append(Cell(routine_link, "keyAddress"))
            row.append(Cell(purpose, "purpose"))
            table.add_row(row)
        ofile.write(table.html())

        self.write_footer(ofile)

    def get_playarea_tile_text(self, tile):
        text = ['<i>Q_addr</i>=(x,181)=%i, <i>Q</i>=%i' % (tile.q_addr, self.skool.snapshot[tile.q_addr])]
        text.append('<i>ref_addr</i>=(<i>Q</i>,y+160)=%i' % tile.ref_addr)
        text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
        text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
        text.append('<i>attr</i>=%i' % tile.attr)
        return text

def join(*path_components):
    return '/'.join([c for c in path_components if c.replace('/', '')])

class FileInfo:
    sd_skool = 'sd.skool'
    bts_skool = 'bts.skool'
    sd_ref = 'sd.ref'
    bts_ref = 'bts.ref'
    skool_ref = 'skool.ref'
    start_skool = 'start.skool'
    save_skool = 'save.skool'
    load_skool = 'load.skool'
    changelog_ref = 'changelog.ref'

    sd_dir = 'skool_daze'
    bts_dir = 'back_to_skool'
    changelog = 'changelog.html'
    home = 'index.html'

    css = 'style.css'
    js = 'skool.js'
    asm_path = 'asm'
    lessons_path = 'lessons'
    images_path = 'images'

    index = 'index.html'

    logo = join(images_path, 'logo.png')
    bubble_img = join(images_path, 'bubble.png')
    font = join(images_path, 'font.png')

    load_path = 'load-save'
    load_index = join(load_path, 'load.html')
    save = join(load_path, 'save.html')

    maps_path = 'maps'
    memory_map = join(maps_path, 'all.html')
    routines = join(maps_path, 'routines.html')
    data = join(maps_path, 'data.html')
    messages = join(maps_path, 'messages.html')

    graphics_path = 'graphics'
    playarea = join(graphics_path, 'playarea.html')
    patiles_path = join(graphics_path, 'patiles')
    playarea_tiles = join(patiles_path, 'patiles.html')
    asstart = join(graphics_path, 'asstart.html')
    animatory_states = join(graphics_path, 'as.html')
    astiles_path = join(graphics_path, 'astiles')
    animatory_state_tiles = join(astiles_path, 'astiles.html')
    graphics = join(graphics_path, 'graphics.html')
    graphic_glitches = join(graphics_path, 'glitches.html')

    buffers_path = 'buffers'
    gbuffer = join(buffers_path, 'gbuffer.html')
    cbuffer = join(buffers_path, 'cbuffer.html')
    personal_timetables = join(lessons_path, 'timetables.html')
    lesson_index = join(lessons_path, 'index.html')
    keys = join('tables', 'keys.html')

    reference_path = 'reference'
    glossary = join(reference_path, 'glossary.html')
    facts = join(reference_path, 'facts.html')
    bugs = join(reference_path, 'bugs.html')
    pokes = join(reference_path, 'pokes.html')

    def __init__(self, topdir, game_dir):
        self.odir = join(topdir, game_dir)

    def open_file(self, *names):
        path = self.odir
        for name in names:
            path = join(path, name)
        if not isdir(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        return open(path, 'wb')

    def file_exists(self, fname):
        return isfile(join(self.odir, fname))

    @classmethod
    def relpath(cls, cwd, target):
        # Use os.path.relpath(target, cwd) in Python 2.6
        while cwd and cwd[-1] == '/':
            cwd = cwd[:-1]
        cwddirs = cwd.split('/')
        while cwddirs and not cwddirs[0]:
            cwddirs.pop(0)
        targetdirs = target.split('/')
        while cwddirs and cwddirs[0] == targetdirs[0]:
            cwddirs.pop(0)
            targetdirs.pop(0)
        return ('../' * len(cwddirs)) + '/'.join(targetdirs)

    @classmethod
    def asm_fname(cls, address, path=''):
        return join(path, '%s.html' % address)

    @classmethod
    def lesson_fname(cls, lesson_num):
        return '%s.html' % lesson_num

    @classmethod
    def css_relpath(cls, cwd):
        return cls.relpath(cwd, '../style.css')

    @classmethod
    def js_relpath(cls, cwd):
        return cls.relpath(cwd, '../skool.js')

    @classmethod
    def asm_relpath(cls, cwd, address):
        return cls.relpath(cwd, join(cls.asm_path, cls.asm_fname(address)))

    @classmethod
    def lesson_relpath(cls, cwd, lesson_num):
        return cls.relpath(cwd, join(cls.lessons_path, cls.lesson_fname(lesson_num)))

    @classmethod
    def load_relpath(cls, cwd, address):
        return cls.relpath(cwd, join(cls.load_path, cls.asm_fname(address)))

    @classmethod
    def patile_fname(cls, page, ref, attr, scale):
        suffix = '' if scale == 4 else 'x%i' % scale
        return join(cls.patiles_path, '%x%x%s.png' % (ref + 256 * page, attr, suffix))

    @classmethod
    def as_fname(cls, state, attr, scale, mask, snapshot_name):
        suffix = '_%s' % snapshot_name if snapshot_name else ''
        fname = 'as%s_%ix%i%s%s.png' % (str(state).zfill(3), attr, scale, 'm' if mask else 'u', suffix)
        return join(cls.images_path, 'animatory_states', fname)

    @classmethod
    def astile_fname(cls, page, ref, attr, scale, mask):
        attr_infix = '' if attr == 120 else '_%x' % attr
        scale_infix = '' if scale == 4 else 'x%i' % scale
        suffix = '' if mask else 'u'
        return join(cls.astiles_path, '%x%s%s%s.png' % (ref + 256 * page, attr_infix, scale_infix, suffix))

    @classmethod
    def udg_fname(cls, address, attr, scale):
        return join(cls.images_path, 'udgs', 'udg%s_%sx%s.png' % (address, attr, scale))

    @classmethod
    def udg_matrix_fname(cls, basename, attr, scale):
        return join(cls.images_path, 'udgs', '%s_%sx%s.png' % (basename, attr, scale))

    @classmethod
    def box_fname(cls, address, attr, scale, snapshot_name):
        suffix = '_%s' % snapshot_name if snapshot_name else ''
        attr_infix = '_%ix' % attr if attr else 'x'
        fname = 'box%i%s%i%s.png' % (address, attr_infix, scale, suffix)
        return join(cls.images_path, 'boxes', fname)

    @classmethod
    def mutable_fname(cls, address, padding):
        fname = 'mutable%s_%s.png' % (address, padding) if padding else 'mutable%s.png' % address
        return join(cls.images_path, 'mutables', fname)

    @classmethod
    def scr_fname(cls, skoolfile):
        return join(cls.images_path, skoolfile[:skoolfile.index('.')] + ".png")

    @classmethod
    def skoolimg_fname(cls, x, y, w, h, scale, show_chars, show_x, snapshot_name):
        suffix = '_%s' % snapshot_name if snapshot_name else ''
        chars_infix = 'c' if show_chars else 'e'
        fname = 'skool%ix%ix%ix%ix%i_%s%i%s.png' % (x, y, w, h, scale, chars_infix, show_x, suffix)
        return join(cls.images_path, 'skool', fname)

class Table:
    def __init__(self, cls='', *headers):
        self.cls = cls
        self.width = None
        self.rows = []
        if headers:
            self.rows.append([Header(h) for h in headers])
        self.cellpadding = None

    def add_row(self, row):
        self.rows.append(row)

    def html(self):
        lines = []
        cls_attr = ' class="%s"' % self.cls if self.cls else ''
        width_attr = ' width="%i"' % self.width if self.width else ''
        cp_attr = ' cellpadding="%i"' % self.cellpadding if self.cellpadding is not None else ''
        lines.append('<table%s%s%s>' % (width_attr, cls_attr, cp_attr))
        for row in self.rows:
            lines.append('<tr>')
            for cell in row:
                lines.append(cell.html())
            lines.append('</tr>')
        lines.append('</table>\n')
        return '\n'.join(lines)

class Cell:
    def __init__(self, contents, cls='', colspan=1, rowspan=1, tag='td'):
        self.contents = str(contents)
        self.cls = cls
        self.colspan = colspan
        self.rowspan = rowspan
        self.tag = tag

    def html(self):
        cls_attr = ' class="%s"' % self.cls if self.cls else ''
        cs_attr = ' colspan="%s"' % self.colspan if self.colspan > 1 else ''
        rs_attr = ' rowspan="%s"' % self.rowspan if self.rowspan > 1 else ''
        if self.contents:
            return '<%s%s%s%s>%s</%s>' % (self.tag, cls_attr, cs_attr, rs_attr, self.contents, self.tag)
        return '<%s%s%s%s/>' % (self.tag, cls_attr, cs_attr, rs_attr)

class Header(Cell):
    def __init__(self, contents, cls='', colspan=1, rowspan=1):
        Cell.__init__(self, contents, cls, colspan, rowspan, 'th')
